# -*- coding: utf-8 -*-

"""Definition of the redis locking backend."""

from __future__ import absolute_import


try:
    from urlparse import urlparse, parse_qsl
except:
    # Python 3!
    from urllib.parse import urlparse, parse_qsl

from celery_once.helpers import now_unix
from celery_once.tasks import AlreadyQueued


def parse_url(url):
    """
    Parse the argument url and return a redis connection.
    Two patterns of url are supported:
        * redis://host:port[/db][?options]
        * redis+socket:///path/to/redis.sock[?options]
    A ValueError is raised if the URL is not recognized.
    """
    parsed = urlparse(url)
    kwargs = parse_qsl(parsed.query)

    # TCP redis connection
    if parsed.scheme == 'redis':
        details = {'host': parsed.hostname}
        if parsed.port:
            details['port'] = parsed.port
        if parsed.password:
            details['password'] = parsed.password
        db = parsed.path.lstrip('/')
        if db and db.isdigit():
            details['db'] = db

    # Unix socket redis connection
    elif parsed.scheme == 'redis+socket':
        details = {'unix_socket_path': parsed.path}
    else:
        raise ValueError('Unsupported protocol %s' % (parsed.scheme))

    # Add kwargs to the details and convert them to the appropriate type, if needed
    details.update(kwargs)
    if 'socket_timeout' in details:
        details['socket_timeout'] = float(details['socket_timeout'])
    if 'db' in details:
        details['db'] = int(details['db'])

    return details


class Redis(object):
    """Redis backend."""

    def __init__(self, settings):
        try:
            from redis import StrictRedis
        except ImportError:
            raise ImportError(
                "You need to install the redis library in order to use Redis"
                " backend (pip install redis)")
        self._redis = StrictRedis(**parse_url(settings['url']))

    @property
    def redis(self):
        # Used to allow easy mocking when testing.
        return self._redis

    def raise_or_lock(self, key, timeout):
        """
        Checks if the task is locked and raises an exception, else locks
        the task.
        """
        now = now_unix()
        # Check if the tasks is already queued if key is in cache.
        result = self.redis.get(key)
        if result:
            # Work out how many seconds remaining till the task timeout.
            remaining = int(result) - now
            if remaining > 0:
                raise AlreadyQueued(remaining)

        # By default, the tasks and the key expire after 60 minutes.
        # (meaning it will not be executed and the lock will clear).
        self.redis.setex(key, timeout, now + timeout)

    def clear_lock(self, key):
        """Remove the lock from redis."""
        return self.redis.delete(key)




from celery import task
from celery_once.tasks import QueueOnce


@task(name='simple_example', base=QueueOnce)
def simple_example():
    return "simple"


@task(name='bound_task', bind=True, base=QueueOnce)
def bound_task(self, a, b):
    return a + b


@task(name='args_example', base=QueueOnce)
def args_example(a, b):
    return a + b


@task(name='select_args_example', base=QueueOnce, once={'keys': ['a']})
def select_args_example(a, b):
    return a + b


def test_get_key_simple():
    assert "qo_simple_example" == simple_example.get_key()


def test_get_key_args_1():
    assert "qo_args_example_a-1_b-2" == args_example.get_key(
        kwargs={'a': 1, 'b': 2})


def test_get_key_args_2():
    assert "qo_args_example_a-1_b-2" == args_example.get_key(args=(1, 2, ))


def test_get_key_select_args_1():
    assert "qo_select_args_example_a-1" == select_args_example.get_key(
        kwargs={'a': 1, 'b': 2})


def test_get_key_bound_task():
    assert "qo_bound_task_a-1_b-2" == bound_task.get_key(
        kwargs={'a': 1, 'b': 2})



import pytest
from freezegun import freeze_time
from fakeredis import FakeStrictRedis

from celery import Celery
from celery_once import QueueOnce, AlreadyQueued


@pytest.fixture()
def redis(monkeypatch):
    fake_redis = FakeStrictRedis()
    fake_redis.flushall()
    monkeypatch.setattr("celery_once.backends.redis.Redis.redis", fake_redis)
    return fake_redis


app = Celery()
app.conf.ONCE = {
    'backend': "celery_once.backends.redis.Redis",
    'settings': {
        'url': "redis://localhost:1337/0",
        'timeout': 30 * 60
    }
}
app.conf.CELERY_ALWAYS_EAGER = True


@app.task(name="example", base=QueueOnce)
def example(a=1):
    assert example.once_backend.redis.get("qo_example_a-1") is not None


def test_init():
    details = example.once_backend.redis.connection_pool.connection_kwargs
    assert details['host'] == "localhost"
    assert details['port'] == 1337
    assert details['db'] == 0


def test_delay(redis):
    example.delay(1)
    assert redis.get("qo_example_a-1") is None


def test_delay_already_queued(redis):
    redis.set("qo_example_a-1", 10000000000)
    try:
        example.delay(1)
        pytest.fail("Didn't raise AlreadyQueued.")
    except AlreadyQueued:
        pass


@freeze_time("2012-01-14")  # Time since epoch = 1326499200
def test_delay_expired(redis):
    # Fallback, key should of been timed out.
    redis.set("qo_example_a-1", 1326499200 - 60 * 60)
    example.delay(1)
    assert redis.get("qo_example_a-1") is None


def test_apply_async(redis):
    example.apply_async(args=(1, ))
    assert redis.get("qo_example_a-1") is None


def test_apply_async_queued(redis):
    redis.set("qo_example_a-1", 10000000000)
    try:
        example.apply_async(args=(1, ))
        pytest.fail("Didn't raise AlreadyQueued.")
    except AlreadyQueued:
        pass


def test_already_queued_graceful(redis):
    redis.set("qo_example_a-1", 10000000000)
    result = example.apply_async(args=(1, ), once={'graceful': True})
    assert result.result is None


@freeze_time("2012-01-14")  # Time since epoch = 1326499200
def test_apply_async_expired(redis):
    # Fallback, key should of been timed out.
    redis.set("qo_example_a-1", 1326499200 - 60 * 60)
    example.apply_async(args=(1, ))



# -*- coding: utf-8 -*-
from celery_once.helpers import (
    queue_once_key, kwargs_to_list, force_string, import_backend)

import pytest
import six


def test_force_string_1():
    assert force_string('a') == 'a'


def test_force_string_2():
    assert force_string(u'a') == 'a'


def test_force_string_3():
    assert force_string('é') == 'é'


def test_force_string_4():
    assert force_string(u'é') == 'é'


def test_kwargs_to_list_empty():
    keys = kwargs_to_list({})
    assert keys == []


def test_kwargs_to_list_1():
    keys = kwargs_to_list({'int': 1})
    assert keys == ["int-1"]


def test_kwargs_to_list_2():
    keys = kwargs_to_list({'int': 1, 'boolean': True})
    assert keys == ["boolean-True", "int-1"]


def test_kwargs_to_list_3():
    keys = kwargs_to_list({'int': 1, 'boolean': True, 'str': "abc"})
    assert keys == ["boolean-True", "int-1", "str-abc"]


def test_kwargs_to_list_4():
    keys = kwargs_to_list(
        {'int': 1, 'boolean': True, 'str': 'abc', 'list': [1, '2']})
    assert keys == ["boolean-True", "int-1", "list-[1, '2']", "str-abc"]


@pytest.mark.skipif(six.PY3, reason='requires python 2')
def test_kwargs_to_list_5():
    keys = kwargs_to_list(
        {'a': {u'é': 'c'}, 'b': [u'a', 'é'], u'c': 1, 'd': 'é', 'e': u'é'})
    assert keys == [
        "a-{'\\xc3\\xa9': 'c'}",
        "b-['a', '\\xc3\\xa9']",
        "c-1",
        "d-\xc3\xa9",
        "e-\xc3\xa9",
    ]


@pytest.mark.skipif(six.PY2, reason='requires python 3')
def test_kwargs_to_list_6():
    keys = kwargs_to_list(
        {'a': {u'é': 'c'}, 'b': [u'a', 'é'], u'c': 1, 'd': 'é', 'e': u'é'})
    assert keys == ["a-{'é': 'c'}", "b-['a', 'é']", "c-1", "d-é", 'e-é']


def test_queue_once_key():
    key = queue_once_key("example", {})
    assert key == "qo_example"


def test_queue_once_key_kwargs():
    key = queue_once_key("example", {'pk': 10})
    assert key == "qo_example_pk-10"


def test_queue_once_key_kwargs_restrict_keys():
    key = queue_once_key("example", {'pk': 10, 'id': 10}, restrict_to=['pk'])
    assert key == "qo_example_pk-10"


@pytest.mark.skipif(six.PY3, reason='requires python 2')
def test_queue_once_key_unicode_py2():
    key = queue_once_key(u"éxample", {'a': u'é', u'b': 'é'})
    assert key == "qo_\xc3\xa9xample_a-\xc3\xa9_b-\xc3\xa9"


@pytest.mark.skipif(six.PY2, reason='requires python 3')
def test_queue_once_key_unicode_py3():
    key = queue_once_key(u"éxample", {'a': u'é', u'b': 'é'})
    assert key == "qo_éxample_a-é_b-é"

class TestBackend(object):
    def __init__(self, settings):
        self.settings = settings


def test_import_backend():
    config = {
        'backend': "tests.backends.TestBackend",
        'settings': 1
    }
    backend = import_backend(config)
    assert backend.settings == 1



import pytest
from freezegun import freeze_time
from fakeredis import FakeStrictRedis

from celery_once.backends.redis import parse_url, Redis
from celery_once.tasks import AlreadyQueued


def test_parse_redis_details_tcp_default_args():
    details = parse_url('redis://localhost:6379/')
    assert details == {'host': 'localhost', 'port': 6379}


def test_parse_url_tcp_with_db():
    details = parse_url('redis://localhost:6379/3')
    assert details == {'host': 'localhost', 'port': 6379, 'db': 3}


def test_parse_url_tcp_no_port():
    details = parse_url('redis://localhost')
    assert details == {'host': 'localhost'}


def test_parse_url_tcp_with_password():
    details = parse_url('redis://:ohai@localhost:6379')
    assert details == {'host': 'localhost', 'port': 6379, 'password': 'ohai'}


def test_parse_url_unix_sock_no_options():
    details = parse_url('redis+socket:///var/run/redis/redis.sock')
    assert details == {'unix_socket_path': '/var/run/redis/redis.sock'}


def test_parse_url_unix_sock_with_options():
    details = parse_url('redis+socket:///var/run/redis/redis.sock?db=2&socket_timeout=2')
    assert details == {
        'unix_socket_path': '/var/run/redis/redis.sock',
        'db': 2,
        'socket_timeout': 2.0
    }


def test_parse_unsupported_url():
    with pytest.raises(ValueError):
        parse_url('amqp://guest:guest@localhost:5672/potato')


@pytest.fixture()
def redis(monkeypatch):
    fake_redis = FakeStrictRedis()
    fake_redis.flushall()
    monkeypatch.setattr("celery_once.backends.redis.Redis.redis", fake_redis)
    return fake_redis


@pytest.fixture()
def backend():
    backend = Redis({'url': "redis://localhost:1337"})
    return backend


@freeze_time("2012-01-14")  # Time since epoch = 1326499200
def test_redis_raise_or_lock(redis, backend):
    assert redis.get("test") is None
    backend.raise_or_lock(key="test", timeout=60)
    assert redis.get("test") is not None


@freeze_time("2012-01-14")  # Time since epoch = 1326499200
def test_redis_raise_or_lock_locked(redis, backend):
    # Set to expire in 30 seconds!
    redis.set("test", 1326499200 + 30)
    with pytest.raises(AlreadyQueued) as e:
        backend.raise_or_lock(key="test", timeout=60)
    assert e.value.countdown == 30
    assert e.value.message == "Expires in 30 seconds"


@freeze_time("2012-01-14")  # Time since epoch = 1326499200
def test_redis_raise_or_lock_locked_and_expired(redis, backend):
    # Set to have expired 30 ago seconds!
    redis.set("test", 1326499200 - 30)
    backend.raise_or_lock(key="test", timeout=60)
    assert redis.get("test") is not None


def test_redis_clear_lock(redis, backend):
    redis.set("test", 1326499200 + 30)
    backend.clear_lock("test")
    assert redis.get("test") is None



import pytest
import mock

from celery import Celery
from celery_once import QueueOnce, AlreadyQueued


app = Celery()
app.conf.ONCE = {
    'backend': 'tests.backends.TestBackend',
    'settings': {
        'timeout': 60
    }
}
app.conf.CELERY_ALWAYS_EAGER = True


@pytest.fixture(autouse=True)
def mock_backend(mocker):
    mocker.patch('tests.backends.TestBackend.raise_or_lock')
    mocker.patch('tests.backends.TestBackend.clear_lock')


@app.task(name="example", base=QueueOnce)
def example():
    return

@app.task(name="example_unlock_before_run", base=QueueOnce, once={'unlock_before_run': True})
def example_unlock_before_run():
    return


def test_config():
    assert example.config == app.conf


def test_once_config():
    assert example.once_config == {
        'backend': 'tests.backends.TestBackend',
        'settings': {
            'timeout': 60
        }
    }


def test_default_timeout():
    assert example.default_timeout == 60


def test_apply_async():
    example.apply_async()
    example.once_backend.raise_or_lock.assert_called_with(
        "qo_example", timeout=60)


def test_apply_async_timeout(mocker):
    example.once_backend.raise_or_lock = mocker.Mock()
    example.apply_async(once={'timeout': 120})
    example.once_backend.raise_or_lock.assert_called_with(
        "qo_example", timeout=120)


def test_raise_already_queued():
    example.once_backend.raise_or_lock.side_effect = AlreadyQueued(60)
    with pytest.raises(AlreadyQueued):
        example.apply_async()


def test_raise_already_queued_graceful():
    example.once_backend.raise_or_lock.side_effect = AlreadyQueued(60)
    result = example.apply_async(once={'graceful': True})
    assert result.result is None


def test_delay_unlock_before_run(mocker):
    mock_parent = mocker.Mock()
    clear_lock_mock = mocker.Mock()
    after_return_mock = mocker.Mock()
    mock_parent.attach_mock(clear_lock_mock, 'clear_lock')
    mock_parent.attach_mock(after_return_mock, 'after_return')
    example_unlock_before_run.once_backend.clear_lock.side_effect = clear_lock_mock
    example_unlock_before_run.after_return = after_return_mock
    example_unlock_before_run.apply_async()
    assert len(mock_parent.mock_calls) == 2
    assert mock_parent.mock_calls[0] == mocker.call.clear_lock('qo_example_unlock_before_run')
